/* Program TRANSFER.CC  */
/* By Harald Stauss     */
/* Last change 20.08.98 */

#include <ctype.h>
#include "tsa.h"
#include "nrc/nrutil.h" 
#include "nrc/nrutil.c"
#include "nrc/gammln.c"
#include "nrc/gser.c"
#include "nrc/gcf.c"
#include "nrc/gammq.c"
#include "nrc/fitul.c"
#include "nrc/four1.c"
#include "nrc/complex.c"


void cmoveavg(float d[], unsigned long n, short s)

  /* Given a complex vector d[1..2*n], and given the number of complex      */
  /* elements n, as well as a smoothing factor s, this routine returns      */
  /* the moveing average over the vector d with (2 * s +1) elements         */
  /* smoothed.                                                              */

{
  unsigned long  i;
  int   j,sn;
  float sr,si;
  float *h;

  h=(float *)calloc(2*n+1,sizeof(float));

  h[1]=d[1]; h[2]=d[2];
  for (i=2; i<=n; i++) {
    sr=0; si=0; sn=0;
    for (j=-s; j<=s; j++) {
      /*       if ((j+i)<2) { */
      /*          sr+=d[3]; */
      /*          si+=d[4]; */
      /*          }; */
      if ((j+i)>=2 && (j+i)<=n) {
	sr+=d[(j+i)*2-1];
	si+=d[(j+i)*2];
	sn++;
      };
      /*       if ((j+i)>n) { */
      /*          sr+=d[n*2-1]; */
      /*          si+=d[n*2]; */
      /*          }; */
    };
    h[i*2-1]=sr/sn;
    h[i*2]  =si/sn;
  };

  for (i=1; i<=n; i++) {
    d[i*2-1]=h[i*2-1];
    d[i*2]  =h[i*2];
  };

  free(h);
}



main(int argc, char *argv[])
{
           int      i, smooth=4;
	   float    sr, sr1, sr2;
  unsigned long     n, ni;
	   float    *xdata, *sig;
	   float    a, b, siga, sigb, chi2, q;
	   float    *data1, *data2, *sxx, *syy, *sxy;
	   float    mag, pha, coh;
           char     *fin1, *ftin1, *pftin1;
	   char     *fin2, *ftin2, *pftin2;
	   char     *fout, *ftout, sform[40];
	   ofstream datout;
           REGIST   rin1, rin2;
           fcomplex zh, zxx, zyy, zxy;


  /* H e l l o   w o r l d */
  cout << "TRANSFER Ver.1.1 by H. Stauss, 1998 \n";

  /* A r r a y s   a l l o c i e r e n */
  fin1=  new char [255];
  ftin1= new char [255];
  fin2=  new char [255];
  ftin2= new char [255];
  fout=  new char [255];
  ftout= new char [255];

  /* P a r a m e t e r   e i n l e s e n */
  if (argc!=6 && argc!=7)
    {
      cout << "Number of parameters wrong! \n";
      cout << "transfer f_in1 f_in2 f_out n smooth [sr] \n";
      cout << "calculates the magnitude and phase of the transfer function \n";
      cout << "as well as the coherence between f_in1 and f_in2 \n";
      cout << "based on n values and a smoothing factor of smooth \n";
      cout << "zero pads to n values if necessary, \n";
      cout << "f_in1 : name of input file 1 (*.tsa, *.asc). \n";
      cout << "f_in2 : name of input file 2 (*.tsa, *.asc). \n";
      cout << "f_out : name of output file (*_n.asc). \n";
      cout << "n     : number of data values for FFT \n";
      cout << "sr    : sr if ASCII file \n";
      cout << "smooth: smoothing factor for weighing the spectra \n";
      cout << "output: frequency, magnitude, coherence, phase \n";
      goto ende;
    }
  strcpy(fin1,argv[1]); strcpy(ftin1,fin1);
  strcpy(fin2,argv[2]); strcpy(ftin2,fin2);
  strcpy(fout,argv[3]); strcpy(ftout,fout);
  pftin1=strrchr(ftin1,'.');
  for (i=0; pftin1[i]; i++) pftin1[i]=toupper(pftin1[i]);
  pftin2=strrchr(ftin2,'.');
  for (i=0; pftin2[i]; i++) pftin2[i]=toupper(pftin2[i]);

  sscanf(argv[4],"%d",&n);
  sscanf(argv[5],"%d",&smooth);

  cout << "Input  file 1: " << fin1 << "\n";
  cout << "Input  file 2: " << fin2 << "\n";
  cout << "Output file  : " << fout << "\n";
  cout << "use " << n << " values for calculations \n";

  /* E i n g a b e d a t e i   o e f f n e n */
  if (strcmp(pftin1,".TSA")==0) {
    rin1.ReadTSA(fin1);
    sr1=rin1.ch[0].sr;
  }
  if (strcmp(pftin1,".ASC")==0 && argc!=7) {
    cout << "sampling rate must be given if ASCII file \n";
    goto ende;
  }
  if (strcmp(pftin1,".ASC")==0) {
    rin1.ReadASCII(fin1,1);
    sscanf(argv[6], "%f", &sr1);
  }
  if ((strcmp(pftin1,".TSA")!=0) && (strcmp(pftin1,".ASC")!=0)) {
    cout << "File type for input not recognized! \n";
    goto ende;
  }
  printf("Number of channels in file 1: %d  \n",rin1.n);

  if (strcmp(pftin2,".TSA")==0) {
    rin2.ReadTSA(fin2);
    sr2=rin2.ch[0].sr;
  }
  if (strcmp(pftin2,".ASC")==0 && argc!=7) {
    cout << "sampling rate must be given if ASCII file \n";
    goto ende;
  }
  if (strcmp(pftin2,".ASC")==0) {
    rin2.ReadASCII(fin2,1);
    sscanf(argv[6], "%f", &sr2);
  }
  if ((strcmp(pftin2,".TSA")!=0) && (strcmp(pftin2,".ASC")!=0)) {
    cout << "File type for input not recognized! \n";
    goto ende;
  }
  printf("Number of channels in file 2: %d  \n",rin2.n);

  if (rin1.n != rin2.n) {
    cout << "number of channels must be the same in both input files! \n";
    goto ende;
  }

  if (sr1!=sr2) {
    cout << "sampling rate must be the same for fin1 and fin2 ! \n";
    goto ende;
  }
  sr=sr1;

  /* M A I N L O O P  */
  data1 = new float [n];
  data2 = new float [n];
  xdata = new float [n];
  sig   = new float [n];
  for (ni=0; ni<n; ni++) {xdata[ni]=ni; sig[ni]=0; }
  sxx   = new float [2*n];
  syy   = new float [2*n];
  sxy   = new float [2*n];
  for (i=0; i < rin1.n; i++) {
    cout << "file 1 channel " << i << ": " << rin1.ch[i].n;
    /* zero padding both files if necessary */
    if (n>rin1.ch[i].n) {
      /* zero padding 1 */
      cout << " zero padding ! \n";
      /* regressionsgerade berechnen */
      fitul(xdata-1, rin1.ch[i].data-1, rin1.ch[i].n, sig-1, 0,
	    &a, &b, &siga, &sigb, &chi2, &q);
      for (ni=0; ni<rin1.ch[i].n; ni++) data1[ni]=rin1.ch[i].data[ni]-(a+b*ni);
      for (ni=rin1.ch[i].n; ni<n; ni++) data1[ni]=0;
    }
    else {
      /* no zero padding 1 */
      cout << "\n";
      /* regressionsgerade berechnen */
      fitul(xdata-1, rin1.ch[i].data-1, n, sig-1, 0,
	    &a, &b, &siga, &sigb, &chi2, &q);
      for (ni=0; ni<n; ni++) data1[ni]=rin1.ch[i].data[ni]-(a+b*ni);
    }
    cout << "file 2 channel " << i << ": " << rin1.ch[i].n;
    if (n>rin2.ch[i].n) {
      /* zero padding 2 */
      cout << " zero padding ! \n";
      /* regressionsgerade berechnen */
      fitul(xdata-1, rin2.ch[i].data-1, rin2.ch[i].n, sig-1, 0,
	    &a, &b, &siga, &sigb, &chi2, &q);
      for (ni=0; ni<rin2.ch[i].n; ni++) data2[ni]=rin2.ch[i].data[ni]-(a+b*ni);
      for (ni=rin2.ch[i].n; ni<n; ni++) data2[ni]=0;
    }
    else {
      /* no zero padding 2 */
      cout << "\n";
      /* regressionsgerade berechnen */
      fitul(xdata-1, rin2.ch[i].data-1, n, sig-1, 0,
	    &a, &b, &siga, &sigb, &chi2, &q);
      for (ni=0; ni<n; ni++) data2[ni]=rin2.ch[i].data[ni]-(a+b*ni);
    }

    /* calculate sxx, syy, sxy */
    for (ni=0; ni<n; ni++) {
      sxx[2*ni]=data1[ni];
      sxx[2*ni+1]=0;
      syy[2*ni]=data2[ni];
      syy[2*ni+1]=0;
    }
    four1(sxx-1,n,1);
    four1(syy-1,n,1);

    /* cross spectrum berechnen */
    for (ni=0; ni<n; ni++) {
      sxy[ni*2]  =(sxx[ni*2]*syy[ni*2])+(sxx[ni*2+1]*syy[ni*2+1]);
      sxy[ni*2+1]=(syy[ni*2]*sxx[ni*2+1])-(sxx[ni*2]*syy[ni*2+1]);
    };

    /* Power der Eingabedateien */
    for (ni=0; ni<n; ni++) {
      sxx[ni*2]  =(sxx[ni*2]*sxx[ni*2])+(sxx[ni*2+1]*sxx[ni*2+1]);
      sxx[ni*2+1]=0;
      syy[ni*2]  =(syy[ni*2]*syy[ni*2])+(syy[ni*2+1]*syy[ni*2+1]);
      syy[ni*2+1]=0;
    };
    cmoveavg(sxy-1,n,smooth);
    cmoveavg(sxx-1,n,smooth);
    cmoveavg(syy-1,n,smooth);

    sprintf(sform,"_%d.asc",i);
    strcpy(ftout,fout); strcat(ftout,sform);
    datout.open(ftout);
    for (ni=0; ni<n; ni++) {
      coh=(sxy[ni*2]*sxy[ni*2]+sxy[ni*2+1]*sxy[ni*2+1])/(sxx[ni*2]*syy[ni*2]);
      pha=atan2(sxy[ni*2+1],sxy[ni*2]);

      zxy=Complex(sxy[2*ni],sxy[2*ni+1]);
      zxx=Complex(sxx[2*ni],0);
      zyy=Complex(syy[2*ni],0);
      zh=Cdiv(zxy, zxx);
      mag=Cabs(zh);

      datout << ni*sr/n << "\t" << mag << "\t" << coh << "\t" << pha << "\n";
    }
    datout.close();
  }
  delete [] data1; delete [] data2; delete [] xdata; delete [] sig;
  delete [] sxx; delete [] syy; delete [] sxy;

 ende:
  delete [] fin1; delete [] ftin1;
  delete [] fin2; delete [] ftin2;
  delete [] fout; delete [] ftout;
  return(0);
} /* Ende von main */
